"""
Sitemap tree widget.

This renders a tree of the sitemap and updates the node info widget when a node is highlighted.

When a new target is set from the SearchBar, the tree is updated with the new sitemap data.
"""

import webbrowser
from typing import ClassVar

from rich.color import Color
from rich.style import Style
from rich.text import Text
from textual import work
from textual.app import ComposeResult
from textual.containers import ScrollableContainer
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Static, Tree
from textual.widgets.tree import TreeNode

from antur.utils.sitemap_parser import Entry, Error, SitemapParser

from .node_info import NodeInfo


def dict_to_tree(dictionary: dict, tree: Tree | TreeNode) -> None:
    """Move a dictionary into the provided tree structure in-place."""
    for key, value in dictionary.items():
        if isinstance(value, Entry | Error):
            tree.add_leaf(key, data=value)
        else:
            sub_tree = tree.add(key)
            dict_to_tree(value, sub_tree)


class CustomTree(Tree):
    """Custom tree widget to aid with formatting."""

    def render_label(
        self: "CustomTree",
        node: TreeNode,
        base_style: Style,
        additional_style: Style,
    ) -> Text:
        """Render the label with a custom style."""
        if hasattr(node, "data"):  # noqa: SIM102
            if isinstance(node.data, Error):
                additional_style = Style.chain(
                    additional_style,
                    Style.from_color(Color.parse("red")),
                )

        return super().render_label(node, base_style, additional_style)


class SitemapTree(Widget):
    """Sitemap tree widget."""

    target = reactive(None, recompose=True, always_update=True)
    tree_data = reactive({}, recompose=True, always_update=True)

    BINDINGS: ClassVar = [("o", "open", "Open the selected item in the browser.")]

    def __init__(
        self: "SitemapTree",
        url: str | None = None,
        max_concurrent_requests: int = 40,
        *args: tuple,
        **kwargs: dict,
    ) -> None:
        """Initialize the sitemap tree."""
        super().__init__(*args, **kwargs)

        self.target = url
        self.max_concurrent_requests = max_concurrent_requests

    def compose(self: "SitemapTree") -> ComposeResult:
        """Compose the sitemap tree."""
        with ScrollableContainer():
            if isinstance(self.tree_data, Error):
                yield Static(f"Error: {self.tree_data.message}")
                return

            if self.target:
                t = CustomTree(f"Sitemap {self.target}")
                dict_to_tree(self.tree_data, t.root)
                t.root.expand()
                yield t
            else:
                yield Static("Enter a URL above to start.")

    @work(exclusive=True)
    async def watch_target(self: "SitemapTree", value: str) -> None:
        """Watch the target for changes and update the tree."""
        if value:
            self.loading = True

            parser = SitemapParser(value, self.max_concurrent_requests)
            self.tree_data = await parser.parse()

            self.loading = False

            self.notify(
                f"Found {parser.found_urls} URLs and {parser.found_sitemaps} sitemaps.",
                title="Search Complete",
            )

            if parser.http_errors:
                self.notify(
                    f"Failed to fetch {parser.http_errors} URLs.",
                    title="HTTP Errors",
                    severity="error",
                )

            if parser.xml_errors:
                self.notify(
                    f"Failed to parse {parser.xml_errors} XML documents.",
                    title="XML Errors",
                    severity="error",
                )

    def on_tree_node_highlighted(self: "SitemapTree", message: Tree.NodeHighlighted) -> None:
        """Update the node info when a node is highlighted."""
        if hasattr(message.node, "data"):
            self.app.query_one(NodeInfo).node = message.node.data
        else:
            self.app.query_one(NodeInfo).node = None

    def action_open(self: "SitemapTree") -> None:
        """Open the selected item in the browser."""
        tree = self.query_one(Tree)
        node = tree.get_node_at_line(tree.cursor_line)

        if node:
            if node.data:
                webbrowser.open(node.data.url)
            else:
                webbrowser.open(str(node.label))



"""Main module for the Antur application."""

import argparse

from . import __version__
from .app import AnturApp


def parse_args() -> argparse.Namespace:
    """Parse the command line arguments."""
    parser = argparse.ArgumentParser(
        description="Antur allows you to browse sitemaps from your command line.",
        epilog=f"antur v{__version__}",
    )
    parser.add_argument("url", nargs="?", help="The URL of a sitemap to start with.")

    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")

    parser.add_argument(
        "-c",
        "--concurrent",
        type=int,
        help="The number of concurrent requests to make.",
        default=40,
    )
    return parser.parse_args()


def main() -> None:
    """Run the Antur application."""
    args = parse_args()

    app = AnturApp(args.url, max_concurrent_requests=args.concurrent)
    app.run()


if __name__ == "__main__":
    main()



"""Parse a sitemap file and return a dict of URLs and their metadata."""

import asyncio
from dataclasses import dataclass

import aiohttp
from lxml.etree import Element, XMLSyntaxError, fromstring, tostring

from antur import __version__


@dataclass
class Entry:
    """Dataclass for a sitemap entry."""

    url: str
    lastmod: str
    changefreq: str
    priority: str
    other: str


@dataclass
class Error:
    """Dataclass for an error."""

    url: str
    message: str


HEADERS = {
    "User-Agent": f"Mozilla/5.0 (compatible; AnturSitemap/{__version__}; +http://github.com/jb3/antur)",
}

IGNORE_TAGS = ["lastmod", "changefreq", "priority", "loc"]


class SitemapParser:
    """Parse a sitemap file and return a dict of URLs and their metadata."""

    def __init__(self: "SitemapParser", url: str, max_concurrent_requests: int) -> None:
        """Initialize the parser."""
        self.url = url
        self.data = {}
        self.found_urls = 0
        self.found_sitemaps = 0
        self.http_errors = 0
        self.xml_errors = 0

        self.semaphore = asyncio.Semaphore(max_concurrent_requests)

    async def get_data(self: "SitemapParser", url: str) -> bytes:
        """Fetch the data from the URL."""
        async with (
            aiohttp.ClientSession() as session,
            session.get(
                url,
                headers=HEADERS,
            ) as response,
        ):
            return await response.read()

    def _filter_out_children(self: "SitemapParser", element: Element) -> Element:
        """Return an XML Element with matching children removed."""
        for tag in IGNORE_TAGS:
            found = element.findall(f"{{*}}{tag}")
            for c in found:
                c.getparent().remove(c)

        return element

    def _maybe_child(self: "SitemapParser", element: Element, tag: str) -> str | None:
        """Return the text of a child element if it exists."""
        child = element.find(tag)
        return child.text if child is not None else None

    async def parse(self: "SitemapParser", url: str | None = None) -> dict[str, Entry | Error]:
        """Parse the sitemap at the given URL."""
        if not url:
            url = self.url

        level = {}

        async with self.semaphore:
            try:
                data = await self.get_data(url)
            except aiohttp.ClientError as e:
                self.http_errors += 1
                return Error(url, str(e))

        try:
            parsed = fromstring(data)  # noqa: S320
        except XMLSyntaxError as e:
            self.xml_errors += 1
            return Error(url, str(e))

        if parsed.tag.endswith("sitemapindex"):
            child_tasks = {}
            for child in parsed:
                if child.tag.endswith("sitemap"):
                    loc = child.find("{*}loc").text
                    self.found_sitemaps += 1
                    child_tasks[loc] = self.parse(loc)

            results = await asyncio.gather(*child_tasks.values())

            level_data = dict(zip(child_tasks.keys(), results, strict=False))

            level.update(level_data)

        if parsed.tag.endswith("urlset"):
            for child in parsed:
                loc = child.find("{*}loc").text
                self.found_urls += 1

                other_formatted = tostring(self._filter_out_children(child), pretty_print=True)

                level[loc] = Entry(
                    loc,
                    self._maybe_child(child, "{*}lastmod"),
                    self._maybe_child(child, "{*}changefreq"),
                    self._maybe_child(child, "{*}priority"),
                    other_formatted.decode(),
                )

        return level



"""
Node info widget.

Shows information on the selected node including the last modified date, change frequency,
and priority.

Optionally shows the markdown preview of the XML content.
"""

from datetime import datetime

from humanize import naturaldelta
from textual.app import ComposeResult
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Markdown

from antur.utils.sitemap_parser import Error

URL_TEMPLATE = """
# {node.url}

- Last Modified: {node.lastmod} {delta}
- Change Frequency: {node.changefreq}
- Priority: {node.priority}

{opt_md}
"""

ERROR_TEMPLATE = """
# Error: {node.message}

Open this URL in your browser to see the error, it is possible that the URL is not a sitemap.
"""


class NodeInfo(Widget):
    """Node info widget."""

    node = reactive(None, recompose=True, always_update=True)

    def compose(self: "NodeInfo") -> ComposeResult:
        """Compose the node info."""
        if self.node:
            if isinstance(self.node, Error):
                yield Markdown(ERROR_TEMPLATE.format(node=self.node), id="node-info")
                return

            try:
                lastmod = datetime.fromisoformat(self.node.lastmod)

                delta = naturaldelta(datetime.now(tz=lastmod.tzinfo) - lastmod)
                delta = f"**({delta} ago)**"
            except (TypeError, ValueError):
                delta = ""

            opt_md = ("```xml\n" + self.node.other + "\n```") if self.app.show_markdown else ""

            yield Markdown(
                URL_TEMPLATE.format(node=self.node, delta=delta, opt_md=opt_md),
                id="node-info",
            )
        else:
            yield Markdown("# No node selected.")



"""Widget for containing the search bar and buttons."""

from urllib.parse import urlparse

from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.validation import Function
from textual.widget import Widget
from textual.widgets import Button, Input

from .sitemap_tree import SitemapTree


def is_url(value: str) -> bool:
    """Check if the value is a URL."""
    try:
        result = urlparse(value)
        if result.scheme not in ["http", "https"]:
            return False

        return all([result.scheme, result.netloc])
    except ValueError:
        return False


class SearchBar(Widget):
    """Search bar widget."""

    def __init__(self: "SearchBar", url: str | None = None, *args: tuple, **kwargs: dict) -> None:
        """Initialize the search bar."""
        super().__init__(*args, **kwargs)

        self.url = url

    def compose(self: "SearchBar") -> ComposeResult:
        """Compose the search bar."""
        with Horizontal(id="buttons-container"):
            yield Input(
                self.url,
                "Sitemap URL",
                id="search-bar",
                validate_on=["changed", "submitted"],
                validators=[Function(is_url, "Value is not a URL.")],
            )
            yield Button("Clear", "error", id="clear-button")
            yield Button("Search", "success", id="search-button")

    def do_search(self: "SearchBar", url: str) -> None:
        """Perform a search."""
        self.notify(url, title="Search Called")
        self.app.query_one(SitemapTree).target = url

    def on_button_pressed(self: "SearchBar", message: Button.Pressed) -> None:
        """Handle button presses."""
        if message.button.id == "clear-button":
            self.query_one(Input).value = ""
        elif message.button.id == "search-button":
            inp = self.query_one(Input)
            if self.query_one(Input).validate(inp.value).is_valid:
                self.do_search(inp.value)
            else:
                self.notify("Invalid URL", title="Search Error", severity="error")

    def on_input_submitted(self: "SearchBar", message: Input.Submitted) -> None:
        """Handle input submission."""
        if message.validation_result.is_valid:
            self.notify(message.value, title="Search Called")
            self.app.query_one(SitemapTree).target = message.value
            self.app.query_one(SitemapTree).refresh()
        else:
            self.notify("Invalid URL", title="Search Error", severity="error")



"""Main application for Antur."""

import typing

from textual.app import App, ComposeResult
from textual.containers import Vertical
from textual.reactive import reactive
from textual.widgets import Footer, Header

from antur import __version__

from .widgets.node_info import NodeInfo
from .widgets.search_bar import SearchBar
from .widgets.sitemap_tree import SitemapTree


class AnturApp(App):
    """Antur allows you to browse sitemaps from your command line."""

    BINDINGS: typing.ClassVar = [
        ("q", "quit", "Quit the application."),
        ("d", "toggle_dark", "Toggle between light and dark mode."),
        (
            "m",
            "toggle_markdown",
            "Show the markdown preview of unrecognised properties.",
        ),
    ]

    CSS_PATH = "antur.tcss"

    ENABLE_COMMAND_PALETTE = False

    TITLE = "Antur"

    SUB_TITLE = f"v{__version__}"

    show_markdown = reactive(default=False)

    def __init__(self: "AnturApp", url: str | None = None, *, max_concurrent_requests: int) -> None:
        """Initialize the Antur application."""
        super().__init__()

        self.url = url
        self.max_concurrent_requests = max_concurrent_requests

    def compose(self: "AnturApp") -> ComposeResult:
        """Compose the layout of the app."""
        yield Header(show_clock=False)
        yield SearchBar(self.url)
        with Vertical(id="contents"):
            with Vertical(id="main"):
                yield SitemapTree(self.url, max_concurrent_requests=self.max_concurrent_requests)
                yield NodeInfo()
            yield Footer()

    def action_toggle_dark(self: "AnturApp") -> None:
        """Toggle between light and dark mode."""
        self.dark = not self.dark

    def action_toggle_markdown(self: "AnturApp") -> None:
        """Toggle the markdown preview."""
        self.show_markdown = not self.show_markdown
        self.query_one(NodeInfo).node = self.query_one(NodeInfo).node
